;;; File: "jss-compile.scm"

;;; Author: Marc Feeley
;;;   Dec 4, 2007 : version 1.0

;;; Modified by: Catherine Gaudron
;;;  last modification: Jan 4, 2008

;;;----------------------------------------------------------------------------

;;; Compiler entry point.

(define (compile expr cont)
  (scheme->js (cps (expand (expand-macro expr)) cont)))

;;;----------------------------------------------------------------------------

;;; Pattern matching.

(define-macro (match sujet . clauses)

  (define (if-equal? var gab oui non)
    (cond ((and (pair? gab)
                (eq? (car gab) 'unquote)
                (pair? (cdr gab))
                (null? (cddr gab)))
           `(let ((,(cadr gab) ,var))
              ,oui))
          ((null? gab)
           `(if (null? ,var) ,oui ,non))
          ((symbol? gab)
           `(if (eq? ,var ',gab) ,oui ,non))
          ((or (boolean? gab)
               (char? gab))
           `(if (eq? ,var ,gab) ,oui ,non))
          ((number? gab)
           `(if (eqv? ,var ,gab) ,oui ,non))
          ((pair? gab)
           (let ((carvar (gensym))
                 (cdrvar (gensym)))
             `(if (pair? ,var)
                  (let ((,carvar (car ,var)))
                    ,(if-equal?
                      carvar
                      (car gab)
                      `(let ((,cdrvar (cdr ,var)))
                         ,(if-equal?
                           cdrvar
                           (cdr gab)
                           oui
                           non))
                      non))
                  ,non)))
          (else
           (error "unknown pattern"))))

  (let* ((var
          (gensym))
         (fns
          (map (lambda (x) (gensym))
               clauses))
         (err
          (gensym)))
    `(let ((,var ,sujet))
       ,@(map (lambda (fn1 fn2 clause)
                `(define (,fn1)
                   ,(if-equal? var
                               (car clause)
                               (if (and (eq? (cadr clause) 'when)
                                        (pair? (cddr clause)))
                                   `(if ,(caddr clause)
                                        ,(cadddr clause)
                                        (,fn2))
                                   (cadr clause))
                               `(,fn2))))
              fns
              (append (cdr fns) (list err))
              clauses)
       (define (,err) (error "match failed"))
       (,(car fns)))))

;;;----------------------------------------------------------------------------

;;; Macro expander.

;; Generates an (env . expr) pair where env has all macros defined in the
;; top level of the initial expression and where expr is the initial expression
;; from which we have removed the macro definition
(define (get-env e)
  (if (null? e) '(())
      (let* ((result (get-env (cdr e)))
             (newenv (car result))
             (newexp (cdr result)))
        (cond
         ((and (list? (car e)) (eq? (caar e) 'define-macro))
          ;; If we have a macro definition
          (let ((name (cadar e))
                (body (cddar e)))
            (if (pair? name)
                (cons (cons (cons (car name)
                                  `(lambda ,(cdr name) ,@body)) newenv)
                      newexp)
                (cons (cons name body)
                      newexp))))
         ;; regular expression
         (else (cons newenv (cons (car e) newexp)))))))

;; Macro expander, for now the only place a macro-definition can
;; be placed is in a begin
(define (expand-macro e)
  (define (exp-macro env e)
    (match e
       (,s when (or (self-evaluating? s) (symbol? s))
           s)
           
       ((begin . ,es)
        (let* ((result (get-env es))
               (newenv (append (car result) env))
               (newexp (cdr result))
               (newes (map (lambda (exp) (exp-macro newenv exp)) newexp)))
          `(begin ,@newes)))

       ((quote ,e1)
        e)
       
       ((,e1 . ,es)
        (let ((macro (assoc e1 env)))
          (if macro
              (if (list? es)
                  (eval `(,(cdr macro) ,@(map (lambda (e) `(quote ,e)) es)))
                  (eval `(,(cdr macro) (quote ,es))))
              (if (list? es)
                  (map (lambda (exp) (exp-macro env exp)) e)
                  (cons (exp-macro env e1) (exp-macro env es))))))
       (,_ e)))
  (exp-macro '() e))

;;expansion of special forms

(define (expand expr)
  (define (expand-quasiquote e)
    (if (pair? e)
        (cond
         ((eq? (car e) 'unquote)
          (expand (cadr e)))
         ((and (pair? (car e)) (eq? (caar e) 'unquote-splicing))
          (list 'append (expand (cadar e)) (expand-quasiquote (cdr e))))
         (else (list 'cons
                     (expand-quasiquote (car e))
                     (expand-quasiquote (cdr e)))))
        (list 'quote e)))
         
  (match expr

    (,c when (constant? c)
     c)

    (,v when (symbol? v)
     v)

    ((quote ,x)
     `(quote ,x))

    ((quasiquote ,x)
     (expand-quasiquote x))

    ((set! ,v ,E1)
     `(set! ,v ,(expand E1)))
    ((set! . ,rest)
     (error "improper set!"))

    ((if ,E1 ,E2)
     `(if ,(expand E1) ,(expand E2) #f))
    ((if ,E1 ,E2 ,E3)
     `(if ,(expand E1) ,(expand E2) ,(expand E3)))
    ((if . ,rest)
     (error "improper if"))

    ((begin ,E1 . ,Es)
     `(begin ,@(map expand (cons E1 Es))))
    ((begin . ,Es)
     (error "improper begin"))

    ((lambda ,params . ,Es)
     `(lambda ,params ,(expand `(begin ,@Es))))
    ((lambda . ,rest)
     (error "improper lambda"))

    ((let ,name ,bindings . ,Es) when (symbol? name)
     (expand
      `((letrec ((,name
                  (lambda ,(map car bindings) ,@Es)))
          ,name)
        ,@(map cadr bindings))))
    ((let ,bindings . ,Es)
     (expand
      `((lambda ,(map car bindings) ,@Es)
        ,@(map cadr bindings))))
    ((let . ,rest)
     (error "improper let"))

    ((letrec ,bindings . ,Es)
     (expand
      `(let ,(map (lambda (b) `(,(car b) #f)) bindings)
         ,@(map (lambda (b) `(set! ,(car b) ,(cadr b))) bindings)
         ,@Es)))
    ((letrec . ,rest)
     (error "improper letrec"))

    ((define (,v . ,params) . ,Es)
     (expand
      `(set! ,v (lambda ,params ,@Es))))
    ((define ,v ,E1)
     (expand
      `(set! ,v ,E1)))
    ((define . ,rest)
     (error "improper define"))

    ((cond)
     `#f)
    ((cond (else ,E1 . ,Es))
     (expand
      `(begin ,E1 ,@Es)))
    ((cond (else . ,Es) . ,rest)
     (error "improper else clause"))
    ((cond (,test) . ,rest)
     (expand
      `(or ,test (cond ,@rest))))
    ((cond (,test => ,fn) . ,rest)
     (expand
      (let ((v (gensym)))
        `(let ((,v ,test))
           (if ,v
               (,fn ,v)
               (cond ,@rest))))))
    ((cond (,test => . ,Es) . ,rest)
     (error "improper => clause"))
    ((cond (,test ,E1 . ,Es) . ,rest)
     (expand
      `(if ,test
           (begin ,E1 ,@Es)
           (cond ,@rest))))
    ((cond . ,rest)
     (error "improper cond"))

    ((case ,key . ,seqs)
     (expand
      (let* ((v (gensym))
             (thunk-name (map (lambda (x) (gensym)) seqs))
             (thunk-list
              (map (lambda (name seq) `(,name (lambda () ,(cadr seq))))
                   thunk-name seqs))
             (cond-list
              (map (lambda (name seq)
                     (if (eq? (car seq) 'else)
                         `(else (,name))
                         `((memv ,v (quote ,(car seq))) (,name))))
                   thunk-name seqs)))
        `(let ((,v ,key)
               ,@thunk-list)
           (cond ,@cond-list)))))


    ((or)
     `#f)
    ((or ,E1)
     (expand E1))
    ((or ,E1 ,E2 . ,Es)
     (expand
      (let ((v (gensym)))
        `(let ((,v ,E1))
           (if ,v ,v (or ,E2 ,@Es))))))

    ((and)
     `#t)
    ((and ,E1)
     (expand E1))
    ((and ,E1 ,E2 . ,Es)
     (expand
      `(if ,E1 (and ,E2 ,@Es) #f)))

    ((,op . ,Es) when (primitive? op)
     `(,op ,@(map expand Es)))

    ((,E0 . ,Es)
     (map expand (cons E0 Es)))

    (,_
     (error "unknown expression" expr))))

(define (constant? c)
  (match c
    ((quote ,x)
     #t)
    (,x
     (self-evaluating? x))))

(define (self-evaluating? obj)
  (or (number? obj)
      (string? obj)
      (boolean? obj)
      (char? obj)))

;;;----------------------------------------------------------------------------

;;; CPS conversion.

(define (cps E K)

  (define (simple? expr)
    (match expr
      (,c when (constant? c)
       #t)
      (,v when (symbol? v)
       #t)
      ((lambda ,params ,E0)
       #t)
      ((set! ,v ,E1)
       (simple? E1))
      ((,op . ,Es) when (primitive? op)
       (all simple? Es))
      (,_
       #f)))

  (define (all pred lst)
    (or (null? lst)
        (and (pred (car lst))
             (all pred (cdr lst)))))

  (define (simple-cps expr)
    (match expr
      (,c when (constant? c)
       c)
      (,v when (symbol? v)
       v)
      ((set! ,v ,E1)
       `(set! ,v ,(simple-cps E1)))
      ((,op . ,Es) when (primitive? op)
       `(,op ,@(map simple-cps Es)))
      ((lambda ,params ,E0)
       (let ((k (gensym)))
         `(lambda (,k ,@params) ,(cps E0 k))))))

  (define (multi-cps exprs body)

    (define (inner sexpr)
      (multi-cps (cdr exprs)
                 (lambda (sexprs)
                   (body (cons sexpr sexprs)))))

    (if (null? exprs)
        (body '())
        (let ((expr (car exprs)))
          (if (simple? expr)
              (inner (simple-cps expr))
              (let ((r (gensym)))
                (cps expr
                     `(lambda (,r)
                        ,(inner r))))))))

  (match E

    (,s when (simple? s)
     `(,K ,(simple-cps s)))

    ((set! ,v ,E1)
     (multi-cps (list E1)
                (lambda (sexprs)
                  `(,K (set! ,v ,@sexprs)))))

    ((if ,E1 ,E2 ,E3)
     (if (not (symbol? K))
         (let ((k (gensym)))
           `((lambda (,k) ,(cps E k)) ,K))
         (multi-cps (list E1)
                    (lambda (sexprs)
                      `(if ,@sexprs
                           ,(cps E2 K)
                           ,(cps E3 K))))))

    ((begin ,E1 . ,Es)
     (if (null? Es)
         (cps E1 K)
         (let ((r1 (gensym)))
           (cps E1 `(lambda (,r1)
                      ,(cps `(begin ,@Es) K))))))

    ((,op . ,Es) when (primitive? op)
     (multi-cps Es
                (lambda (sexprs)
                  `(,K (,op ,@sexprs)))))

    ((,E0 . ,Es)
     (multi-cps (cons E0 Es)
                (lambda (sexprs)
                  `(,(car sexprs)
                    ,K
                    ,@(cdr sexprs)))))

    (,_
     (error "unknown expression"))))

;;;----------------------------------------------------------------------------

;;; JavaScript code generation.

(define (scheme->js exprcps)
  (set! rev-inits '(" "))
  (let ((js (scm->js exprcps)))
    (list
     "function () { "
     (reverse rev-inits)
     " return "
     js
     "; }")))

(define (scm->js exprcps)

  (define (variable->string sym)
    (if (jss-function-name? sym)
        (symbol->string sym)
        (string-append
         "scm_"
         (apply string-append
                (map (lambda (c)
                       (if (or (char-alphabetic? c)
                               (char-numeric? c))
                           (string c)
                           (string-append
                            "_"
                            (number->string (char->integer c) 16)
                            "_")))
                     (string->list (symbol->string sym)))))))

  (define (restparam->var args);;FIXME no need to append
    (let loop ((args args) (n 0))
      (if (pair? args)
          (string-append "var " (variable->string (car args))
                         " = arguments[" (number->string n) "]; "
                         (loop (cdr args) (+ n 1)))
          (let ((var (variable->string args)))
            (string-append "var " var "= null; for(i = arguments.length; i>" 
                           (number->string n)
                           ";--i){ " var " = jss_cons (arguments[i], "
                           var ")};")))))

  (match exprcps

    (,c when (self-evaluating? c)
     (datum->js c))

    ((quote ,datum)
     (datum->js datum))

    (,v when (symbol? v)
     (variable->string v))

    ((set! ,v ,E1)
     (list
      (variable->string v)
      " = "
      (scm->js E1)))

    ((if ,E1 ,E2 ,E3)
     (list
      (scm->js E1)
      " ? "
      (scm->js E2)
      " : "
      (scm->js E3)))
    
     ((lambda ,params ,E0)
      (let ((E0 (scm->js E0)))
        (if (list? params)
            (list "function (" (separated-list "," (map variable->string params))
                  ") { return " E0 " }")
            (list "function () { " (restparam->var params)
                  "return " E0 " }"))))

    ((,op . ,Es) when (primitive? op)
     (conv-prim op
                (map scm->js Es)))

    ((,E0 . ,Es)
     (list
      "function() { return "
      (scm->js E0)
      "("
      (separated-list "," (map scm->js Es))
      ") }"))

    (,_
     (error "unknown expression" exprcps))))

(define (datum->js obj)
  (define (prebuild js)
    (let ((var (string-append "jss_datum" (number->string datum-count))))
      (set! datum-count (+ datum-count 1))
      (set! rev-inits
            (cons (list var " = " js "; ")
                  rev-inits))
      var))

  (cond ((number? obj)
         (number->string obj))
        ((string? obj)
         (object->string obj))
        ((boolean? obj)
         (if obj "true" "false"))
        ((null? obj)
         "null")
        ((pair? obj)
         (prebuild
          (list "jss_cons ("
                (datum->js (car obj))
                ", "
                (datum->js (cdr obj))
                ")")))
        ((vector? obj)
         (prebuild
          (list "jss_list2vector ("
                (datum->js (vector->list obj))
                ")")))
        ((symbol? obj)
         (prebuild
          (list "jss_string2symbol ("
                (datum->js (symbol->string obj))
                ")")))
        ((char? obj)
         (prebuild
          (list "jss_make_char ("
                (char->integer obj)
                ")")))
        (else
         (error "datum->js can't handle" obj))))

(define rev-inits #f)
(define datum-count 0)

(define (jss-function-name? sym)
  (let ((str (symbol->string sym)))
    (and (>= (string-length str) 4)
         (string=? (substring str 0 4) "jss_"))))

;;;----------------------------------------------------------------------------

;;; Handling of primitives.

(define primitives '(+ - * / = < > <= >=))

(define (primitive? op)
  (and (symbol? op)
       (or (memq op primitives)
           (memv #\. (string->list (symbol->string op)))
           (jss-function-name? op))))

(define (conv-prim op args)
  (let ((nb-args (length args)))
    (case op
      ((+)
       (case nb-args
         ((0) "0")
         ((1) (car args))
         (else (conv-binary-oper "+" args))))
      ((-)
       (case nb-args
         ((0) (error "wrong number of arguments to '-'"))
         ((1) (conv-prefix-oper "-" args))
         (else (conv-binary-oper "-" args))))
      ((*)
       (case nb-args
         ((0) "1")
         ((1) (car args))
         (else (conv-binary-oper "*" args))))
      ((/)
       (case nb-args
         ((0) (error "wrong number of arguments to '/'"))
         ((1) (conv-prefix-oper "1 /" args))
         (else (conv-binary-oper "/" args))))
      ((=)
       (case nb-args
         ((2) (conv-binary-oper "==" args))
         (else (error "wrong number of arguments to '='"))))
      ((<)
       (case nb-args
         ((2) (conv-binary-oper "<" args))
         (else (error "wrong number of arguments to '<'"))))
      ((>)
       (case nb-args
         ((2) (conv-binary-oper ">" args))
         (else (error "wrong number of arguments to '>'"))))
      ((<=)
       (case nb-args
         ((2) (conv-binary-oper "<=" args))
         (else (error "wrong number of arguments to '<='"))))
      ((>=)
       (case nb-args
         ((2) (conv-binary-oper ">=" args))
         (else (error "wrong number of arguments to '>='"))))
      (else
       (list
        (symbol->string op)
        "("
        (separated-list "," args)
        ")")))))

(define (conv-prefix-oper op args)
  (list "(" op " " (car args) ")"))

(define (conv-suffix-oper op args)
  (list "(" (car args) op ")"))

(define (conv-binary-oper op args)
  (list "(" (separated-list (list " " op " ") args) ")"))

(define (separated-list sep elems)
  (if (null? elems) '()
      (if (null? (cdr elems))
          (car elems)
          (list
           (car elems)
           sep
           (separated-list sep (cdr elems))))))

;;;----------------------------------------------------------------------------
